import json
from pathlib import Path
from tempfile import NamedTemporaryFile, TemporaryDirectory
from .config import VERSION as version
from .logger import get_logger
from .bandcamp import Bandcamp, BandcampError
from .media import LocalMedia
from .download import (download_file, unzip_file, move_file, copy_file,
                       mask_sig, is_zip_file, DownloadInvalidContentType,
                       DownloadBadStatusCode)


log = logger.get_logger('sync')


def do_sync(cookies_path, cookies, dir_path, media_format):
    local_media = LocalMedia(media_dir=dir_path)
    bandcamp = Bandcamp(cookies=cookies)
    bandcamp.verify_authentication()
    bandcamp.load_purchases()
    for item in bandcamp.purchases:
        if local_media.is_locally_downloaded(item.item_id):
            log.info(f'Already locally downloaded, skipping: "{item.band_name} / {item.item_title} "'
                     f'(id:{item.item_id})')
            continue
        else:
            log.info(f'New media item, will download: "{item.band_name} / {item.item_title} "'
                     f'(id:{item.item_id}) in "{media_format}"')
            local_path = local_media.get_path_for_purchase(item)
            local_path.mkdir(parents=True, exist_ok=True)
            try:
                initial_download_url = bandcamp.get_download_file_url(item, encoding=media_format)
            except BandcampError as e:
                log.error(f'Failed to locate download URL for media item "{item.band_name} / {item.item_title} " '
                          f'(id:{item.item_id}), unable to download release ({e}), skipping')
                continue
            download_file_url = bandcamp.check_download_stat(item, initial_download_url)
            with NamedTemporaryFile(mode='w+b', delete=True) as temp_file:
                log.info(f'Downloading item "{item.band_name} / {item.item_title}" (id:{item.item_id}) '
                         f'from {mask_sig(download_file_url)} to {temp_file.name}')
                try:
                    download_file(download_file_url, temp_file)
                except DownloadBadStatusCode as e:
                    log.error(f'Download attempt returned an unexpected status code ({e}), skipping')
                    continue
                except DownloadInvalidContentType as e:
                    log.error(f'Download attempt returned an unexpected content type ({e}), skipping')
                    continue
                temp_file.seek(0)
                temp_file_path = Path(temp_file.name)
                if is_zip_file(temp_file_path):
                    with TemporaryDirectory() as temp_dir:
                        log.info(f'Decompressing downloaded zip "{temp_file.name}" to "{temp_dir}"')
                        unzip_file(temp_file.name, temp_dir)
                        temp_path = Path(temp_dir)
                        for file_path in temp_path.iterdir():
                            file_dest = local_media.get_path_for_file(local_path, file_path.name)
                            log.info(f'Moving extracted file: "{file_path}" to "{file_dest}"')
                            move_file(file_path, file_dest)
                    local_media.write_bandcamp_id(item, local_path)
                elif item.item_type == 'track':
                    slug = item.url_hints.get('slug', item.item_title)
                    file_dest = local_media.get_path_for_file(local_path, f'{slug}.{media_format}')
                    log.info(f'Copying single track: "{temp_file_path}" to "{file_dest}"')
                    copy_file(temp_file_path, file_dest)
                    local_media.write_bandcamp_id(item, local_path)
                else:
                    log.error(f'Downloaded file for "{item.band_name} / {item.item_title}" (id:{item.item_id}) '
                              f'at "{temp_file_path}" is not a zip archive or a single track, skipping')
    return True



import json
import re
from time import time
from http.cookies import SimpleCookie
from html import unescape as html_unescape
from urllib.parse import urlsplit, urlunsplit
from bs4 import BeautifulSoup
import requests
from .config import USER_AGENT
from .logger import get_logger


log = get_logger('bandcamp')


class BandcampError(ValueError):
    pass


class Bandcamp:

    BASE_PROTO = 'https'
    BASE_DOMAIN = 'bandcamp.com'
    URLS = {
        'index': '/',
        'collection_items': '/api/fancollection/1/collection_items',
    }

    def __init__(self, cookies=''):
        self.is_authenticated = False
        self.user_id = 0
        self.user_name = ''
        self.user_url = ''
        self.user_verified = False
        self.user_private = False
        self.cookies = SimpleCookie()
        self.purchases = []
        try:
            self.cookies.load(cookies)
        except Exception as e:
            raise BandcampError(f'Failed to parse cookies string: {e}') from e
        session = self.cookies.get('session')
        if not session:
            raise BandcampError(f'Cookie data does not contain a session value, make sure your '
                                f'cookies.txt file is valid and you copied it from an '
                                f'authenticated browser')
        session_snip = session.value[:20]
        log.info(f'Located Bandcamp session in cookies: {session_snip}...')
        # Create a requests session and map our SimpleCookie to it
        self.session = requests.Session()
        for cookie_name, morsel in self.cookies.items():
            self.session.cookies.set(cookie_name, morsel.value)

    @property
    def cookies_str(self):
        return self.cookies.output(header='').strip().replace('\r\n', ';')

    def refresh_cookes(self, file_path):
        log.info(f'Refreshing cookies: {file_path}')
        with open(file_path, 'wt') as f:
            f.write(self.cookies_str)

    def _construct_url(self, url_name):
        if url_name not in self.URLS:
            raise BandcampError(f'URL name is unknown: {url_name}')
        return urlunsplit((self.BASE_PROTO, self.BASE_DOMAIN, self.URLS[url_name], '', ''))

    def _plain_cookies(self):
        cookies = {}
        for (cookie_name, cookie_value) in self.cookies.items():
            cookies[cookie_value.key] = cookie_value.value
        return cookies

    def _request(self, method, url, data=None, json_data=None, is_json=False, as_raw=False):
        headers = {'User-Agent': USER_AGENT}
        try:
            # The debug logs do not mask the URL which may be a security issue if you run
            # with level=logging.DEBUG
            log.debug(f'Making {method} request to {url}')
            response = self.session.request(
                method,
                url,
                headers=headers,
                cookies=self._plain_cookies(),
                data=data,
                json=json_data
            )
        except Exception as e:
            raise BandcampError(f'Failed to make HTTP request to {mask_sig(url)}: {e}') from e
        if response.status_code != 200:
            raise BandcampError(f'Failed to make HTTP request to {mask_sig(url)}: '
                                f'unknown status code response: {response.status_code}')
        if as_raw:
            return response.text
        elif is_json:
            return json.loads(response.text)
        else:
            return BeautifulSoup(response.text, 'html.parser')

    def _extract_pagedata_from_soup(self, soup):
        pagedata_tag = soup.find('div', id='pagedata')
        if not pagedata_tag:
            raise BandcampError(f'Failed to locate <div id="pagedata"> in index HTML, this may '
                                f'be an authentication issue or it may be that bandcamp.com has '
                                f'updated their website and this tool needs to be updated.')
        encoded_pagedata = pagedata_tag.attrs.get('data-blob')
        if not encoded_pagedata:
            raise BandcampError(f'Failed to extract page data, check your cookies are from an ',
                                f'authenticated session')
        pagedata_str = html_unescape(encoded_pagedata)
        try:
            return json.loads(pagedata_str)
        except Exception as e:
            raise BandcampError(f'Failed to parse pagedata as JSON: {e}') from e

    def _extract_pagedata_from_html(self, html):
        """
            Wrapper for _extract_pagedata_from_soup() that can accept HTML rather than a bs4 soup.
        """
        soup = BeautifulSoup(html, 'html.parser')
        return self._extract_pagedata_from_soup(soup)

    def _get_js_stat_url(self, body, download_url):
        """
            Checks the "stat" download URL body, which is in JavaScript, for
            either the OK response or a new updated download URL.
        """
        body = body.strip()
        if body == "var _statDL_result = { result: 'ok'};":
            # Download is OK, original download URL will work
            return download_url
        # Attempt to find the updated download_url in the JavaScript with a hacky regex
        pattern = re.compile('\"([^\"]+)\":\"([^\"]+)\"')
        for k, v in pattern.findall(body):
            if k == 'download_url':
                return v
        # Fallback to the original download URL
        return download_url

    def verify_authentication(self):
        """
            Loads the initial account and session data from a request to the index page
            of bandcamp.com. When properly authenticated an HTML data attribute is present
            that contains account information in an encoded form.
        """
        url = self._construct_url('index')
        soup = self._request('get', url)
        pagedata = self._extract_pagedata_from_soup(soup)
        try:
            identities = pagedata['identities']
        except KeyError as e:
            raise BandcampError(f'Failed to parse pagedata JSON, does not contain an '
                                f'"identities" key') from e
        try:
            fan = identities['fan']
        except KeyError as e:
            raise BandcampError(f'Failed to parse pagedata JSON, does not contain an '
                                f'"identities.fan" key') from e
        try:
            self.user_id = fan['id']
            self.user_name = fan['name']
            self.user_url = fan['url']
            self.user_verified = fan['verified']
            self.user_private = fan['private']
        except KeyError as e:
            raise BandcampError(f'Failed to parse pagedata JSON, "identities.fan" seems '
                                f'invalid: {fan}') from e
        self.is_authenticated = self.user_id > 0
        log.info(f'Loaded page data, session is authenticated for user '
                 f'"{self.user_name}" (user id:{self.user_id}, url:{self.user_url})')
        return True

    def load_purchases(self):
        """
            Loads all purchases on the authenticated account and returns a list of
            purchase data. Each purchase is a dict of data.
        """
        if not self.is_authenticated:
            raise BandcampError(f'Authentication not verified, call load_pagedata() first')
        log.info(f'Loading purchases for "{self.user_name}" (user id:{self.user_id})')
        self.purchases = []
        now = int(time())
        page_ts = 0
        token = f'{now}:{page_ts}:a::'
        per_page = 100
        while(True):
            log.info(f'Requesting {per_page} purchases using token {token}')
            data = {
                'fan_id': self.user_id,
                'count': per_page,
                'older_than_token': token
            }
            url = self._construct_url('collection_items')
            data = self._request('POST', url, json_data=data, is_json=True)
            try:
                items = data['items']
            except KeyError as e:
                raise BandcampError(f'Failed to extract items from collection results page')
            if not items:
                log.info(f'Reached end of items')
                break
            try:
                redownload_urls = data['redownload_urls']
            except KeyError as e:
                raise BandcampError(f'Failed to extract redownload_urls from collection results page')
            for item_data in items:
                sale_item_type = item_data['sale_item_type']
                sale_item_id = item_data['sale_item_id']
                download_url_key = f'{sale_item_type}{sale_item_id}'
                try:
                    download_url = redownload_urls[download_url_key]
                except KeyError:
                    log.error(f'Failed to locate download URL for {item.band_name} / {item.title} '
                              f'(key:{download_url_key}), skipping item...')
                    continue
                item_data['download_url'] = download_url
                item = BandcampItem(item_data)
                token = item.token
                log.info(f'Found item: {item.band_name} / {item.item_title} (id:{item.item_id})')
                self.purchases.append(item)
        log.info(f'Loaded {len(self.purchases)} purchases')
        return True
    
    def get_download_file_url(self, item, encoding='flac'):
        soup = self._request('get', item.download_url)
        pagedata = self._extract_pagedata_from_soup(soup)
        download_url = None
        if not pagedata:
            raise ValueError(f'Either "url" or "pagedata" must be supplied')
        try:
            digital_items = pagedata['digital_items']
        except KeyError as e:
            raise BandcampError(f'Failed to parse pagedata JSON, does not contain an '
                                f'"digital_items" key') from e
        for digital_item in digital_items:
            try:
                digital_item_id = digital_item['item_id']
            except KeyError as e:
                raise BandcampError(f'Failed to parse pagedata JSON, does not contain an '
                                    f'"digital_items[].art_id" key') from e
            if digital_item_id == item.item_id:
                try:
                    downloads = digital_item['downloads']
                except KeyError as e:
                    raise BandcampError(f'Failed to parse pagedata JSON, does not contain an '
                                        f'"digital_items.downloads" key') from e
                try:
                    download_format = downloads[encoding]
                except KeyError as e:
                    encodings = download_format.keys()
                    raise BandcampError(f'Download formats does not contain requested encoding: {encoding} '
                                        f'(available encodings: {encodings})') from e
                try:
                    download_url = download_format['url']
                except KeyError as e:
                    raise BandcampError(f'Failed to parse pagedata JSON, does not contain an '
                                        f'"digital_items.downloads.[encoding].url" key') from e
                return download_url
        return False

    def check_download_stat(self, item, file_download_url):
        """
            Constructs the download "stat" URL and verifies the state of the download.
            If the state is OK, return the existing URL (download is OK) otherwise wait
            for the stat to complete and return the new download URL.
        """
        download_url_parts = urlsplit(file_download_url)
        path = download_url_parts.path
        path_parts = path.split('/')
        if path_parts[1] == 'download':
            path_parts[1] = 'statdownload'
        stat_url = urlunsplit((
            download_url_parts.scheme,
            download_url_parts.netloc,
            '/'.join(path_parts),
            download_url_parts.query,
            ''
        ))
        body = self._request('get', stat_url, as_raw=True)
        return self._get_js_stat_url(body, file_download_url)


class BandcampItem:

    def __init__(self, data):
        self._data = data

    def __repr__(self):
        return json.dumps(self._data, indent=4, sort_keys=True)

    def __getattr__(self, key):
        try:
            return self._data[key]
        except KeyError as e:
            raise KeyError(f'BandcampItem value "{key}" does not exist') from e



from unicodedata import normalize
from .logger import get_logger


log = get_logger('media')


class LocalMedia:
    """
        A local media directory indexer. This stores media in the following format:

            /media_dir/
            /media_dir/Artist Name
            /media_dir/Artist Name/Album Name
            /media_dir/Artist Name/Album Name/bandcamp_item_id.txt
            /media_dir/Artist Name/Album Name/track1.flac
            /media_dir/Artist Name/Album Name/track2.flac
    """

    ITEM_INDEX_FILENAME = 'bandcamp_item_id.txt'

    def __init__(self, media_dir):
        self.media_dir = media_dir
        self.media = {}
        self.dirs = set()
        log.info(f'Local media directory: {self.media_dir}')
        self.index()

    def _clean_path(self, path_str):
        path_str = str(path_str)
        disallowed_punctuation = '"#%\'*/?\\`'
        normalized_path = normalize('NFKD', path_str)
        outstr = ''
        for c in normalized_path:
            if c not in disallowed_punctuation:
                outstr += c
        return outstr

    def index(self):
        for child1 in self.media_dir.iterdir():
            if child1.is_dir():
                for child2 in child1.iterdir():
                    if child2.is_dir():
                        self.dirs.add(child2)
                        for child3 in child2.iterdir():
                            if child3.name == self.ITEM_INDEX_FILENAME:
                                item_id = self.read_item_id(child3)
                                self.media[item_id] = child2
                                log.info(f'Detected locally downloaded media: {item_id} = {child2}')
        return True

    def read_item_id(self, filepath):
        with open(filepath, 'rt') as f:
            item_id = f.read().strip()
        try:
            return int(item_id)
        except Exception as e:
            raise ValueError(f'Failed to cast item ID from {filepath} "{item_id}" as an int: {e}') from e

    def is_locally_downloaded(self, item_id):
        return item_id in self.media

    def is_dir(self, path):
        return path in self.dirs

    def get_path_for_purchase(self, item):
        return self.media_dir / self._clean_path(item.band_name) / self._clean_path(item.item_title)

    def get_path_for_file(self, local_path, file_name):
        return local_path / self._clean_path(file_name)

    def write_bandcamp_id(self, item, dirpath):
        outfile = dirpath / self.ITEM_INDEX_FILENAME
        log.info(f'Writing bandcamp item id:{item.item_id} to: {outfile}')
        with open(outfile, 'wt') as f:
            f.write(f'{item.item_id}\n')
        return True



import os
import sys
from setuptools import setup, find_packages


version = '0.3.1'


with open('README.md', 'rt') as f:
    long_description = f.read()


with open('requirements.txt', 'rt') as f:
    requirements = tuple(f.read().split())


setup(
    name = 'bandcampsync',
    version = version,
    url = 'https://github.com/meeb/bandcampsync',
    author = 'https://github.com/meeb',
    author_email = 'meeb@meeb.org',
    description = 'A Python module and script to synchronise media purchased on bandcamp.com with a local directory.',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    license = 'BSD',
    include_package_data = True,
    install_requires = requirements,
    packages = find_packages(),
    scripts = [
        'bin/bandcampsync',
        'bin/bandcampsync-service',
    ],
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    keywords = ['bandcampsync', 'bandcamp', 'media', 'sync']
)



import math
import shutil
from zipfile import ZipFile
import requests
from .config import USER_AGENT
from .logger import get_logger


log = get_logger('download')


def mask_sig(url):
    if '&sig=' not in url:
        return url
    url_parts = url.split('&')
    for i, url_part in enumerate(url_parts):
        if url_part[:4] == 'sig=':
            url_parts[i] = 'sig=[masked]'
        elif url_part[:6] == 'token=':
            url_parts[i] = 'token=[masked]'
    return '&'.join(url_parts)


class DownloadBadStatusCode(ValueError):
    pass


class DownloadInvalidContentType(ValueError):
    pass


def download_file(url, target, mode='wb', chunk_size=8192, logevery=10, disallow_content_type='text/html'):
    """
        Attempts to stream a download to an open target file handle in chunks. If the
        request returns a disallowed content type then return a failed state with the
        response content.
    """
    text = True if 't' in mode else False
    data_streamed = 0
    last_log = 0
    headers = {'User-Agent': USER_AGENT}
    with requests.get(url, stream=True, headers=headers) as r:
        r.raise_for_status()
        if r.status_code != 200:
            raise DownloadBadStatusCode(f'Got non-200 status code: {r.status_code}')
        try:
            content_type = r.headers.get('Content-Type', '')
        except (ValueError, KeyError):
            content_type = ''
        content_type_parts = content_type.split(';')
        major_content_type = content_type_parts[0].strip()
        if major_content_type == disallow_content_type:
            raise DownloadInvalidContentType(f'Invalid content type: {major_content_type}')
        try:
            content_length = int(r.headers.get('Content-Length', '0'))
        except (ValueError, KeyError):
            content_length = 0
        for chunk in r.iter_content(chunk_size=chunk_size):
            data_streamed += len(chunk)
            if text:
                chunk = chunk.decode()
            target.write(chunk)
            if content_length > 0 and logevery > 0:
                percent_complete = math.floor((data_streamed / content_length) * 100)
                if percent_complete % logevery == 0 and percent_complete > last_log:
                    log.info(f'Downloading {mask_sig(url)}: {percent_complete}%')
                    last_log = percent_complete
    return True


def is_zip_file(file_path):
    try:
        with ZipFile(file_path) as z:
            z.infolist()
        return True
    except Exception as e:
        return False


def unzip_file(decompress_from, decompress_to):
    with ZipFile(decompress_from) as z:
        z.extractall(decompress_to)
    return True


def move_file(src, dst):
    return shutil.move(src, dst)


def copy_file(src, dst):
    return shutil.copyfile(src, dst)
